// **************************************************************************************
// el Comandante                                                    guerilla.bplaced.net
// --------------------------------------------------------------------------------------
// Information:
// Ultrasonic distance measurement, working with different sensor models
// --------------------------------------------------------------------------------------
// Used Libraries:
// TFT_eSPI --> https://github.com/Bodmer/TFT_eSPI
// NewPing  --> https://bitbucket.org/teckel12/arduino-new-ping/wiki/Home
// **************************************************************************************

#include <Arduino.h>                                      //
#include <math.h>                                         //
#include "NewPing.h"                                      // Ultrasonic library
#include <TFT_eSPI.h>                                     // TFT_espi Library
TFT_eSPI tft = TFT_eSPI();
TFT_eSprite needle = TFT_eSprite(&tft);                   // needle sprite

// ######################################################################################
// Change values in this area to match your envoirement and desires
// --------------------------------------------------------------------------------------
// Ultrasonic GPIO
#define TRIG_PIN 25                                       // the IO-pin at which the sensor trigger is connected
#define ECHO_PIN 26                                       // the IO-pin at which the sensor echo is connected
// ######################################################################################

// Vars
uint32_t        update;
uint32_t        width;                                    // tft width
uint32_t        height;                                   // tft height
uint16_t        needleW ;                                 // needle width
uint16_t        needleH ;                                 // needle height
uint32_t        x;
uint32_t        y;
const uint16_t  ultrasonicDistanceMax = 360;              // max. distance
uint16_t        ultrasonicDistanceOld = 1;
uint16_t        ultrasonicDistance = 1;
uint16_t        needleDistance = 0;

// Arrays
int32_t         lineDistanceEndX [360];
int32_t         lineDistanceEndY [360];
int32_t         lineDistanceShortStartX [360];
int32_t         lineDistanceShortStartY [360];
int32_t         lineDistanceLongStartX [360];
int32_t         lineDistanceLongStartY [360];
uint16_t        distanceTextX [360];
uint16_t        distanceTextY [360];
uint16_t        tftPivotX [360];
uint16_t        tftPivotY [360];
uint16_t        needleAngle [ultrasonicDistanceMax];

NewPing sonar(TRIG_PIN, ECHO_PIN, ultrasonicDistanceMax);

// **************************************************************************************
void setup(void)
{
  // configure pins
  pinMode(TFT_BL, OUTPUT);                                // backlight
  pinMode(TRIG_PIN, OUTPUT);                              // ultrasonic trigger
  pinMode(ECHO_PIN, INPUT_PULLDOWN);                      // ultrasonic echo

  // init TFT_espi
  tft.init();

  // build startup screen
  digitalWrite(TFT_BL, 0);                                // switch backlight off
  tft.setRotation(0);
  width = tft.width();
  height = tft.height();
  tft.fillScreen(0x0000);
  tft.setTextSize(1);
  tft.setTextDatum(MC_DATUM);
  scopeCreate();
  needleCreate();
  digitalWrite(TFT_BL, 1);                                // switch tft backlight on
  update = millis();                                      // sensors, next update time
}


// **************************************************************************************
void loop()
{
  if (update <= millis())
  {
    update = millis() + 250;                              // update time in milliseconds
    ultrasonicMeasure();                                  // ultrasonic modul HC-SRF05, distance
    needleDraw();                                         // draw distance needle sprite
  }
}


// **************************************************************************************
void ultrasonicMeasure(void)
{
  ultrasonicDistanceOld = ultrasonicDistance;
  //ultrasonicDistance = random(ultrasonicDistanceMax);
  ultrasonicDistance = sonar.ping_cm();
}


// **************************************************************************************
void needleDraw()
{
  if (ultrasonicDistance != ultrasonicDistanceOld && ultrasonicDistance != 0)
  {
    tft.setTextColor(0xC618, 0x0000);
    tft.setCursor(width / 2 - height * 0.125, height / 2 + width * 0.55);
    tft.printf("%i cm   ", ultrasonicDistance);
    while (needleDistance != ultrasonicDistance)
    {
      if (needleDistance < ultrasonicDistance)
      {needleDistance++;}
      else
      {needleDistance--;}
      x = needleAngle[needleDistance];
      tft.setPivot(tftPivotX[x], tftPivotY[x]);
      needle.pushRotated(x + 45);
    }
  }
}


// **************************************************************************************
void scopeCreate(void)
{
  if (width >= 320 || height >= 320)          // big screen (i.e. WR32-SWC01 - 480 x 320)
  {
    needleW = width / 5.5;                                // needle dimension
    needleH = needleW;
    tft.setTextFont(2);
    // fill arrays, swaping X and Y
    for (x = 0; x < 360; x++)
    {
      // array distance lines
      lineDistanceEndY[x] = height / 2 + width * 0.46 * cos(x * DEG_TO_RAD);
      lineDistanceEndX[x] = width / 2 + width * 0.46 * sin(x * DEG_TO_RAD);
      lineDistanceShortStartY[x] = height / 2 + width * 0.44 * cos(x * DEG_TO_RAD);
      lineDistanceShortStartX[x] = width / 2 + width * 0.44 * sin(x * DEG_TO_RAD);
      lineDistanceLongStartY[x] = height / 2 + width * 0.43 * cos(x * DEG_TO_RAD);
      lineDistanceLongStartX[x] = width / 2 + width * 0.43 * sin(x * DEG_TO_RAD);
      // array distance text
      distanceTextY[x] = height / 2 + width * 0.38 * cos(x * DEG_TO_RAD);
      distanceTextX[x] = width / 2 - width * 0.38 * sin(x * DEG_TO_RAD);
      // tft pivot
      tftPivotY[x] = height / 2 + width * 0.08 * cos(x * DEG_TO_RAD);
      tftPivotX[x] = width / 2 - width * 0.08 * sin(x * DEG_TO_RAD);
    }

    // fill array, precalculated needle angle
    for (x = 0; x < ultrasonicDistanceMax; x++)
    {needleAngle[x] = (float) 30 + 300.0 / ultrasonicDistanceMax * x;}

    // draw distance short lines
    for (x = 30; x <= 330; x = x + 5)
    {tft.drawLine(lineDistanceShortStartX[x], lineDistanceShortStartY[x], lineDistanceEndX[x], lineDistanceEndY[x], 0xFFFE);}

    // draw distance long lines and distance text
    tft.setTextColor(0x9492, 0x0000);
    y = 0;
    for (x = 30; x <= 330; x = x + 20)
    {
      tft.drawWideLine(lineDistanceLongStartX[x], lineDistanceLongStartY[x], lineDistanceEndX[x], lineDistanceEndY[x], 2, 0x0618, 0x0410);
      tft.drawNumber(y, distanceTextX[x], distanceTextY[x], 2);
      y = y + ultrasonicDistanceMax / 15;
    }

    // title text and circle
    tft.fillRoundRect(0, 4, width, 36, 12, 0x1082);
    tft.setTextSize(2);
    tft.setTextColor(0xC618);
    tft.drawString("Distance", width / 2, 22);
    tft.fillCircle(width / 2, height / 2, width * 0.07, 0x4208);
    tft.fillCircle(width / 2, height / 2, width * 0.04, 0x630C);
    tft.drawCircle(width / 2, height / 2, width * 0.07, 0x8410);
  }

  else if (width < 320 || height < 320)       // small screen (i.e. LILYGO TTGO-T-Display - 240 x 135)
  {
    needleW = width / 7;                                  // needle dimension
    needleH = needleW;
    tft.setTextFont(0);
    // fill arrays, swapping X and Y
    for (x = 0; x < 360; x++)
    {
      // array distance lines
      lineDistanceEndY[x] = height / 2 + width * 0.46 * cos(x * DEG_TO_RAD);
      lineDistanceEndX[x] = width / 2 + width * 0.46 * sin(x * DEG_TO_RAD);
      lineDistanceShortStartY[x] = height / 2 + width * 0.44 * cos(x * DEG_TO_RAD);
      lineDistanceShortStartX[x] = width / 2 + width * 0.44 * sin(x * DEG_TO_RAD);
      lineDistanceLongStartY[x] = height / 2 + width * 0.43 * cos(x * DEG_TO_RAD);
      lineDistanceLongStartX[x] = width / 2 + width * 0.43 * sin(x * DEG_TO_RAD);
      // array distance text
      distanceTextY[x] = height / 2 + width * 0.3 * cos(x * DEG_TO_RAD);
      distanceTextX[x] = width / 2 - width * 0.3 * sin(x * DEG_TO_RAD);
      // tft pivot
      tftPivotY[x] = height / 2 + width * 0.02 * cos(x * DEG_TO_RAD);
      tftPivotX[x] = width / 2 - width * 0.02 * sin(x * DEG_TO_RAD);
    }

    // fill array, precalculated needle angle
    for (x = 0; x < ultrasonicDistanceMax; x++)
    {needleAngle[x] = (float) 30 + 300.0 / ultrasonicDistanceMax * x;}

    // draw distance short lines
    for (x = 30; x <= 330; x = x + 5)
    {tft.drawLine(lineDistanceShortStartX[x], lineDistanceShortStartY[x], lineDistanceEndX[x], lineDistanceEndY[x], 0xFFFE);}

    // draw distance long lines and distance text
    y = 0;
    tft.setTextColor(0x9492, 0x0000);
    for (x = 30; x <= 330; x = x + 50)
    {
      tft.drawWideLine(lineDistanceLongStartX[x], lineDistanceLongStartY[x], lineDistanceEndX[x], lineDistanceEndY[x], 2, 0x0618, 0x0410);
      tft.drawNumber(y, distanceTextX[x], distanceTextY[x], 2);
      y = y + ultrasonicDistanceMax / 6;
    }

    // title text and circle
    tft.fillRoundRect(0, 4, width, 28, 12, 0x1082);
    tft.setTextSize(2);
    tft.setTextColor(0xC618);
    tft.drawString(" Distance", width / 2, 18);
    tft.fillCircle(width / 2, height / 2, width * 0.015, 0xFFFE);
    tft.drawCircle(width / 2, height / 2, width * 0.015, 0xFFFE);
  }
}


// **************************************************************************************
void needleCreate(void)
{
  // create needle sprite
  needle.setColorDepth(8);
  needle.createSprite(needleW, needleH);
  needle.fillSprite(0x0000);                              // fill needle with transparent color
  needle.setPivot(0, 0);                                  // pivot point relative to top left corner of Sprite
  //needle.drawRect(0, 0, needleW, needleH, 0x0700);
  needle.drawWideLine(3, 3, needleW - 5, needleH - 5, 3, 0x0618, 0x0410);
  needle.drawWideLine(needleW - 8, needleH - 8, needleW - 5, needleH - 5, 3, 0xF800, 0x0410);
}
